@@ -21,6 +21,7 @@ import (
 	"encoding/json"
 	"errors"
 	"fmt"
+	"hash"
 	"io"
 	"net"
 	"os"
@@ -37,12 +38,16 @@ import (
 //
 // This functionality is EXPERIMENTAL and may be changed in later releases.
 type ObjectStoreManager interface {
-	// ObjectStore will lookup and bind to an existing object store instance.
+	// ObjectStore will look up and bind to an existing object store instance.
 	ObjectStore(bucket string) (ObjectStore, error)
 	// CreateObjectStore will create an object store.
 	CreateObjectStore(cfg *ObjectStoreConfig) (ObjectStore, error)
 	// DeleteObjectStore will delete the underlying stream for the named object.
 	DeleteObjectStore(bucket string) error
+	// ObjectStoreNames is used to retrieve a list of bucket names
+	ObjectStoreNames(opts ...ObjectOpt) <-chan string
+	// ObjectStores is used to retrieve a list of bucket statuses
+	ObjectStores(opts ...ObjectOpt) <-chan ObjectStoreStatus
 }
 
 // ObjectStore is a blob store capable of storing large objects efficiently in
@@ -55,32 +60,32 @@ type ObjectStore interface {
 	// Put will place the contents from the reader into a new object.
 	Put(obj *ObjectMeta, reader io.Reader, opts ...ObjectOpt) (*ObjectInfo, error)
 	// Get will pull the named object from the object store.
-	Get(name string, opts ...ObjectOpt) (ObjectResult, error)
+	Get(name string, opts ...GetObjectOpt) (ObjectResult, error)
 
 	// PutBytes is convenience function to put a byte slice into this object store.
 	PutBytes(name string, data []byte, opts ...ObjectOpt) (*ObjectInfo, error)
 	// GetBytes is a convenience function to pull an object from this object store and return it as a byte slice.
-	GetBytes(name string, opts ...ObjectOpt) ([]byte, error)
+	GetBytes(name string, opts ...GetObjectOpt) ([]byte, error)
 
-	// PutBytes is convenience function to put a string into this object store.
+	// PutString is convenience function to put a string into this object store.
 	PutString(name string, data string, opts ...ObjectOpt) (*ObjectInfo, error)
 	// GetString is a convenience function to pull an object from this object store and return it as a string.
-	GetString(name string, opts ...ObjectOpt) (string, error)
+	GetString(name string, opts ...GetObjectOpt) (string, error)
 
 	// PutFile is convenience function to put a file into this object store.
 	PutFile(file string, opts ...ObjectOpt) (*ObjectInfo, error)
 	// GetFile is a convenience function to pull an object from this object store and place it in a file.
-	GetFile(name, file string, opts ...ObjectOpt) error
+	GetFile(name, file string, opts ...GetObjectOpt) error
 
 	// GetInfo will retrieve the current information for the object.
-	GetInfo(name string) (*ObjectInfo, error)
-	// UpdateMeta will update the meta data for the object.
+	GetInfo(name string, opts ...GetObjectInfoOpt) (*ObjectInfo, error)
+	// UpdateMeta will update the metadata for the object.
 	UpdateMeta(name string, meta *ObjectMeta) error
 
 	// Delete will delete the named object.
 	Delete(name string) error
 
-	// AddLink will add a link to another object into this object store.
+	// AddLink will add a link to another object.
 	AddLink(name string, obj *ObjectInfo) (*ObjectInfo, error)
 
 	// AddBucketLink will add a link to another object store.
@@ -93,7 +98,7 @@ type ObjectStore interface {
 	Watch(opts ...WatchOpt) (ObjectWatcher, error)
 
 	// List will list all the objects in this store.
-	List(opts ...WatchOpt) ([]*ObjectInfo, error)
+	List(opts ...ListObjectsOpt) ([]*ObjectInfo, error)
 
 	// Status retrieves run-time status about the backing store of the bucket.
 	Status() (ObjectStoreStatus, error)
@@ -126,9 +131,20 @@ var (
 	ErrBadObjectMeta        = errors.New("nats: object-store meta information invalid")
 	ErrObjectNotFound       = errors.New("nats: object not found")
 	ErrInvalidStoreName     = errors.New("nats: invalid object-store name")
-	ErrInvalidObjectName    = errors.New("nats: invalid object name")
 	ErrDigestMismatch       = errors.New("nats: received a corrupt object, digests do not match")
+	ErrInvalidDigestFormat  = errors.New("nats: object digest hash has invalid format")
 	ErrNoObjectsFound       = errors.New("nats: no objects found")
+	ErrObjectAlreadyExists  = errors.New("nats: an object already exists with that name")
+	ErrNameRequired         = errors.New("nats: name is required")
+	ErrNeeds262             = errors.New("nats: object-store requires at least server version 2.6.2")
+	ErrLinkNotAllowed       = errors.New("nats: link cannot be set when putting the object in bucket")
+	ErrObjectRequired       = errors.New("nats: object required")
+	ErrNoLinkToDeleted      = errors.New("nats: not allowed to link to a deleted object")
+	ErrNoLinkToLink         = errors.New("nats: not allowed to link to another link")
+	ErrCantGetBucket        = errors.New("nats: invalid Get, object is a link to a bucket")
+	ErrBucketRequired       = errors.New("nats: bucket required")
+	ErrBucketMalformed      = errors.New("nats: bucket malformed")
+	ErrUpdateMetaDeleted    = errors.New("nats: cannot update meta for a deleted object")
 )
 
 // ObjectStoreConfig is the config for the object store.
@@ -136,6 +152,7 @@ type ObjectStoreConfig struct {
 	Bucket      string
 	Description string
 	TTL         time.Duration
+	MaxBytes    int64
 	Storage     StorageType
 	Replicas    int
 	Placement   *Placement
@@ -205,15 +222,14 @@ type ObjectResult interface {
 }
 
 const (
-	objNameTmpl         = "OBJ_%s"
-	objSubjectsPre      = "$O."
-	objAllChunksPreTmpl = "$O.%s.C.>"
-	objAllMetaPreTmpl   = "$O.%s.M.>"
-	objChunksPreTmpl    = "$O.%s.C.%s"
-	objMetaPreTmpl      = "$O.%s.M.%s"
+	objNameTmpl         = "OBJ_%s"     // OBJ_<bucket> // stream name
+	objAllChunksPreTmpl = "$O.%s.C.>"  // $O.<bucket>.C.> // chunk stream subject
+	objAllMetaPreTmpl   = "$O.%s.M.>"  // $O.<bucket>.M.> // meta stream subject
+	objChunksPreTmpl    = "$O.%s.C.%s" // $O.<bucket>.C.<object-nuid> // chunk message subject
+	objMetaPreTmpl      = "$O.%s.M.%s" // $O.<bucket>.M.<name-encoded> // meta message subject
 	objNoPending        = "0"
 	objDefaultChunkSize = uint32(128 * 1024) // 128k
-	objDigestType       = "sha-256="
+	objDigestType       = "SHA-256="
 	objDigestTmpl       = objDigestType + "%s"
 )
 
@@ -226,7 +242,7 @@ type obs struct {
 // CreateObjectStore will create an object store.
 func (js *js) CreateObjectStore(cfg *ObjectStoreConfig) (ObjectStore, error) {
 	if !js.nc.serverMinVersion(2, 6, 2) {
-		return nil, errors.New("nats: object-store requires at least server version 2.6.2")
+		return nil, ErrNeeds262
 	}
 	if cfg == nil {
 		return nil, ErrObjectConfigRequired
@@ -239,16 +255,30 @@ func (js *js) CreateObjectStore(cfg *ObjectStoreConfig) (ObjectStore, error) {
 	chunks := fmt.Sprintf(objAllChunksPreTmpl, name)
 	meta := fmt.Sprintf(objAllMetaPreTmpl, name)
 
+	// We will set explicitly some values so that we can do comparison
+	// if we get an "already in use" error and need to check if it is same.
+	// See kv
+	replicas := cfg.Replicas
+	if replicas == 0 {
+		replicas = 1
+	}
+	maxBytes := cfg.MaxBytes
+	if maxBytes == 0 {
+		maxBytes = -1
+	}
+
 	scfg := &StreamConfig{
 		Name:        fmt.Sprintf(objNameTmpl, name),
 		Description: cfg.Description,
 		Subjects:    []string{chunks, meta},
 		MaxAge:      cfg.TTL,
+		MaxBytes:    maxBytes,
 		Storage:     cfg.Storage,
-		Replicas:    cfg.Replicas,
+		Replicas:    replicas,
 		Placement:   cfg.Placement,
 		Discard:     DiscardNew,
 		AllowRollup: true,
+		AllowDirect: true,
 	}
 
 	// Create our stream.
@@ -260,13 +290,13 @@ func (js *js) CreateObjectStore(cfg *ObjectStoreConfig) (ObjectStore, error) {
 	return &obs{name: name, stream: scfg.Name, js: js}, nil
 }
 
-// ObjectStore will lookup and bind to an existing object store instance.
+// ObjectStore will look up and bind to an existing object store instance.
 func (js *js) ObjectStore(bucket string) (ObjectStore, error) {
 	if !validBucketRe.MatchString(bucket) {
 		return nil, ErrInvalidStoreName
 	}
 	if !js.nc.serverMinVersion(2, 6, 2) {
-		return nil, errors.New("nats: key-value requires at least server version 2.6.2")
+		return nil, ErrNeeds262
 	}
 
 	stream := fmt.Sprintf(objNameTmpl, bucket)
@@ -283,20 +313,22 @@ func (js *js) DeleteObjectStore(bucket string) error {
 	return js.DeleteStream(stream)
 }
 
-func sanitizeName(name string) string {
-	stream := strings.ReplaceAll(name, ".", "_")
-	return strings.ReplaceAll(stream, " ", "_")
+func encodeName(name string) string {
+	return base64.URLEncoding.EncodeToString([]byte(name))
 }
 
 // Put will place the contents from the reader into this object-store.
 func (obs *obs) Put(meta *ObjectMeta, r io.Reader, opts ...ObjectOpt) (*ObjectInfo, error) {
-	if meta == nil {
+	if meta == nil || meta.Name == "" {
 		return nil, ErrBadObjectMeta
 	}
 
-	obj := sanitizeName(meta.Name)
-	if !keyValid(obj) {
-		return nil, ErrInvalidObjectName
+	if meta.Opts == nil {
+		meta.Opts = &ObjectMetaOptions{ChunkSize: objDefaultChunkSize}
+	} else if meta.Opts.Link != nil {
+		return nil, ErrLinkNotAllowed
+	} else if meta.Opts.ChunkSize == 0 {
+		meta.Opts.ChunkSize = objDefaultChunkSize
 	}
 
 	var o objOpts
@@ -309,17 +341,19 @@ func (obs *obs) Put(meta *ObjectMeta, r io.Reader, opts ...ObjectOpt) (*ObjectIn
 	}
 	ctx := o.ctx
 
-	// Grab existing meta info.
-	einfo, err := obs.GetInfo(meta.Name)
+	// Create the new nuid so chunks go on a new subject if the name is re-used
+	newnuid := nuid.Next()
+
+	// These will be used in more than one place
+	chunkSubj := fmt.Sprintf(objChunksPreTmpl, obs.name, newnuid)
+
+	// Grab existing meta info (einfo). Ok to be found or not found, any other error is a problem
+	// Chunks on the old nuid can be cleaned up at the end
+	einfo, err := obs.GetInfo(meta.Name, GetObjectInfoShowDeleted()) // GetInfo will encode the name
 	if err != nil && err != ErrObjectNotFound {
 		return nil, err
 	}
 
-	// Create a random subject prefixed with the object stream name.
-	id := nuid.Next()
-	chunkSubj := fmt.Sprintf(objChunksPreTmpl, obs.name, id)
-	metaSubj := fmt.Sprintf(objMetaPreTmpl, obs.name, obj)
-
 	// For async error handling
 	var perr error
 	var mu sync.Mutex
@@ -334,7 +368,7 @@ func (obs *obs) Put(meta *ObjectMeta, r io.Reader, opts ...ObjectOpt) (*ObjectIn
 		return perr
 	}
 
-	purgePartial := func() { obs.js.purgeStream(obs.stream, &streamPurgeRequest{Subject: chunkSubj}) }
+	purgePartial := func() { obs.js.purgeStream(obs.stream, &StreamPurgeRequest{Subject: chunkSubj}) }
 
 	// Create our own JS context to handle errors etc.
 	js, err := obs.js.nc.JetStream(PublishAsyncErrHandler(func(js JetStream, _ *Msg, err error) { setErr(err) }))
@@ -342,14 +376,11 @@ func (obs *obs) Put(meta *ObjectMeta, r io.Reader, opts ...ObjectOpt) (*ObjectIn
 		return nil, err
 	}
 
-	chunkSize := objDefaultChunkSize
-	if meta.Opts != nil && meta.Opts.ChunkSize > 0 {
-		chunkSize = meta.Opts.ChunkSize
-	}
-
 	m, h := NewMsg(chunkSubj), sha256.New()
-	chunk, sent, total := make([]byte, chunkSize), 0, uint64(0)
-	info := &ObjectInfo{Bucket: obs.name, NUID: id, ObjectMeta: *meta}
+	chunk, sent, total := make([]byte, meta.Opts.ChunkSize), 0, uint64(0)
+
+	// set up the info object. The chunk upload sets the size and digest
+	info := &ObjectInfo{Bucket: obs.name, NUID: newnuid, ObjectMeta: *meta}
 
 	for r != nil {
 		if ctx != nil {
@@ -370,40 +401,45 @@ func (obs *obs) Put(meta *ObjectMeta, r io.Reader, opts ...ObjectOpt) (*ObjectIn
 
 		// Actual read.
 		// TODO(dlc) - Deadline?
-		n, err := r.Read(chunk)
+		n, readErr := r.Read(chunk)
+
+		// Handle all non EOF errors
+		if readErr != nil && readErr != io.EOF {
+			purgePartial()
+			return nil, readErr
+		}
+
+		// Add chunk only if we received data
+		if n > 0 {
+			// Chunk processing.
+			m.Data = chunk[:n]
+			h.Write(m.Data)
+
+			// Send msg itself.
+			if _, err := js.PublishMsgAsync(m); err != nil {
+				purgePartial()
+				return nil, err
+			}
+			if err := getErr(); err != nil {
+				purgePartial()
+				return nil, err
+			}
+			// Update totals.
+			sent++
+			total += uint64(n)
+		}
 
 		// EOF Processing.
-		if err == io.EOF {
-			// Finalize sha.
-			sha := h.Sum(nil)
+		if readErr == io.EOF {
 			// Place meta info.
 			info.Size, info.Chunks = uint64(total), uint32(sent)
-			info.Digest = fmt.Sprintf(objDigestTmpl, base64.URLEncoding.EncodeToString(sha[:]))
+			info.Digest = GetObjectDigestValue(h)
 			break
-		} else if err != nil {
-			purgePartial()
-			return nil, err
 		}
-
-		// Chunk processing.
-		m.Data = chunk[:n]
-		h.Write(m.Data)
-
-		// Send msg itself.
-		if _, err := js.PublishMsgAsync(m); err != nil {
-			purgePartial()
-			return nil, err
-		}
-		if err := getErr(); err != nil {
-			purgePartial()
-			return nil, err
-		}
-		// Update totals.
-		sent++
-		total += uint64(n)
 	}
 
-	// Publish the metadata.
+	// Prepare the meta message
+	metaSubj := fmt.Sprintf(objMetaPreTmpl, obs.name, encodeName(meta.Name))
 	mm := NewMsg(metaSubj)
 	mm.Header.Set(MsgRollup, MsgRollupSubject)
 	mm.Data, err = json.Marshal(info)
@@ -413,7 +449,8 @@ func (obs *obs) Put(meta *ObjectMeta, r io.Reader, opts ...ObjectOpt) (*ObjectIn
 		}
 		return nil, err
 	}
-	// Send meta message.
+
+	// Publish the meta message.
 	_, err = js.PublishMsgAsync(mm)
 	if err != nil {
 		if r != nil {
@@ -426,40 +463,110 @@ func (obs *obs) Put(meta *ObjectMeta, r io.Reader, opts ...ObjectOpt) (*ObjectIn
 	select {
 	case <-js.PublishAsyncComplete():
 		if err := getErr(); err != nil {
-			purgePartial()
+			if r != nil {
+				purgePartial()
+			}
 			return nil, err
 		}
 	case <-time.After(obs.js.opts.wait):
 		return nil, ErrTimeout
 	}
-	info.ModTime = time.Now().UTC()
 
-	// Delete any original one.
+	info.ModTime = time.Now().UTC() // This time is not actually the correct time
+
+	// Delete any original chunks.
 	if einfo != nil && !einfo.Deleted {
-		chunkSubj := fmt.Sprintf(objChunksPreTmpl, obs.name, einfo.NUID)
-		obs.js.purgeStream(obs.stream, &streamPurgeRequest{Subject: chunkSubj})
+		echunkSubj := fmt.Sprintf(objChunksPreTmpl, obs.name, einfo.NUID)
+		obs.js.purgeStream(obs.stream, &StreamPurgeRequest{Subject: echunkSubj})
 	}
 
+	// TODO would it be okay to do this to return the info with the correct time?
+	// With the understanding that it is an extra call to the server.
+	// Otherwise the time the user gets back is the client time, not the server time.
+	// return obs.GetInfo(info.Name)
+
 	return info, nil
 }
 
+// GetObjectDigestValue calculates the base64 value of hashed data
+func GetObjectDigestValue(data hash.Hash) string {
+	sha := data.Sum(nil)
+	return fmt.Sprintf(objDigestTmpl, base64.URLEncoding.EncodeToString(sha[:]))
+}
+
+// DecodeObjectDigest decodes base64 hash
+func DecodeObjectDigest(data string) ([]byte, error) {
+	digest := strings.SplitN(data, "=", 2)
+	if len(digest) != 2 {
+		return nil, ErrInvalidDigestFormat
+	}
+	return base64.URLEncoding.DecodeString(digest[1])
+}
+
 // ObjectResult impl.
 type objResult struct {
 	sync.Mutex
-	info *ObjectInfo
-	r    io.ReadCloser
-	err  error
-	ctx  context.Context
+	info   *ObjectInfo
+	r      io.ReadCloser
+	err    error
+	ctx    context.Context
+	digest hash.Hash
 }
 
 func (info *ObjectInfo) isLink() bool {
 	return info.ObjectMeta.Opts != nil && info.ObjectMeta.Opts.Link != nil
 }
 
+type GetObjectOpt interface {
+	configureGetObject(opts *getObjectOpts) error
+}
+type getObjectOpts struct {
+	ctx context.Context
+	// Include deleted object in the result.
+	showDeleted bool
+}
+
+type getObjectFn func(opts *getObjectOpts) error
+
+func (opt getObjectFn) configureGetObject(opts *getObjectOpts) error {
+	return opt(opts)
+}
+
+// GetObjectShowDeleted makes Get() return object if it was marked as deleted.
+func GetObjectShowDeleted() GetObjectOpt {
+	return getObjectFn(func(opts *getObjectOpts) error {
+		opts.showDeleted = true
+		return nil
+	})
+}
+
+// For nats.Context() support.
+func (ctx ContextOpt) configureGetObject(opts *getObjectOpts) error {
+	opts.ctx = ctx
+	return nil
+}
+
 // Get will pull the object from the underlying stream.
-func (obs *obs) Get(name string, opts ...ObjectOpt) (ObjectResult, error) {
+func (obs *obs) Get(name string, opts ...GetObjectOpt) (ObjectResult, error) {
+	var o getObjectOpts
+	for _, opt := range opts {
+		if opt != nil {
+			if err := opt.configureGetObject(&o); err != nil {
+				return nil, err
+			}
+		}
+	}
+	ctx := o.ctx
+	infoOpts := make([]GetObjectInfoOpt, 0)
+	if ctx != nil {
+		infoOpts = append(infoOpts, Context(ctx))
+	}
+	if o.showDeleted {
+		infoOpts = append(infoOpts, GetObjectInfoShowDeleted())
+	}
+
 	// Grab meta info.
-	info, err := obs.GetInfo(name)
+	info, err := obs.GetInfo(name, infoOpts...)
 	if err != nil {
 		return nil, err
 	}
@@ -467,28 +574,26 @@ func (obs *obs) Get(name string, opts ...ObjectOpt) (ObjectResult, error) {
 		return nil, ErrBadObjectMeta
 	}
 
-	// Check for object links.If single objects we do a pass through.
+	// Check for object links. If single objects we do a pass through.
 	if info.isLink() {
 		if info.ObjectMeta.Opts.Link.Name == _EMPTY_ {
-			return nil, errors.New("nats: link is a bucket")
+			return nil, ErrCantGetBucket
 		}
-		lobs, err := obs.js.ObjectStore(info.ObjectMeta.Opts.Link.Bucket)
+
+		// is the link in the same bucket?
+		lbuck := info.ObjectMeta.Opts.Link.Bucket
+		if lbuck == obs.name {
+			return obs.Get(info.ObjectMeta.Opts.Link.Name)
+		}
+
+		// different bucket
+		lobs, err := obs.js.ObjectStore(lbuck)
 		if err != nil {
 			return nil, err
 		}
 		return lobs.Get(info.ObjectMeta.Opts.Link.Name)
 	}
 
-	var o objOpts
-	for _, opt := range opts {
-		if opt != nil {
-			if err := opt.configureObject(&o); err != nil {
-				return nil, err
-			}
-		}
-	}
-	ctx := o.ctx
-
 	result := &objResult{info: info, ctx: ctx}
 	if info.Size == 0 {
 		return result, nil
@@ -504,7 +609,7 @@ func (obs *obs) Get(name string, opts ...ObjectOpt) (ObjectResult, error) {
 	}
 
 	// For calculating sum256
-	h := sha256.New()
+	result.digest = sha256.New()
 
 	processChunk := func(m *Msg) {
 		if ctx != nil {
@@ -539,24 +644,12 @@ func (obs *obs) Get(name string, opts ...ObjectOpt) (ObjectResult, error) {
 			b = b[n:]
 		}
 		// Update sha256
-		h.Write(m.Data)
+		result.digest.Write(m.Data)
 
 		// Check if we are done.
 		if tokens[ackNumPendingTokenPos] == objNoPending {
 			pw.Close()
 			m.Sub.Unsubscribe()
-
-			// Make sure the digest matches.
-			sha := h.Sum(nil)
-			rsha, err := base64.URLEncoding.DecodeString(info.Digest)
-			if err != nil {
-				gotErr(m, err)
-				return
-			}
-			if !bytes.Equal(sha[:], rsha) {
-				gotErr(m, ErrDigestMismatch)
-				return
-			}
 		}
 	}
 
@@ -572,7 +665,7 @@ func (obs *obs) Get(name string, opts ...ObjectOpt) (ObjectResult, error) {
 // Delete will delete the object.
 func (obs *obs) Delete(name string) error {
 	// Grab meta info.
-	info, err := obs.GetInfo(name)
+	info, err := obs.GetInfo(name, GetObjectInfoShowDeleted())
 	if err != nil {
 		return err
 	}
@@ -580,77 +673,126 @@ func (obs *obs) Delete(name string) error {
 		return ErrBadObjectMeta
 	}
 
-	// Place a rollup delete marker.
+	// Place a rollup delete marker and publish the info
 	info.Deleted = true
 	info.Size, info.Chunks, info.Digest = 0, 0, _EMPTY_
 
-	metaSubj := fmt.Sprintf(objMetaPreTmpl, obs.name, sanitizeName(name))
-	mm := NewMsg(metaSubj)
-	mm.Data, err = json.Marshal(info)
-	if err != nil {
-		return err
-	}
-	mm.Header.Set(MsgRollup, MsgRollupSubject)
-	_, err = obs.js.PublishMsg(mm)
-	if err != nil {
+	if err = publishMeta(info, obs.js); err != nil {
 		return err
 	}
 
 	// Purge chunks for the object.
 	chunkSubj := fmt.Sprintf(objChunksPreTmpl, obs.name, info.NUID)
-	return obs.js.purgeStream(obs.stream, &streamPurgeRequest{Subject: chunkSubj})
+	return obs.js.purgeStream(obs.stream, &StreamPurgeRequest{Subject: chunkSubj})
 }
 
-// AddLink will add a link to another object into this object store.
+func publishMeta(info *ObjectInfo, js JetStreamContext) error {
+	// marshal the object into json, don't store an actual time
+	info.ModTime = time.Time{}
+	data, err := json.Marshal(info)
+	if err != nil {
+		return err
+	}
+
+	// Prepare and publish the message.
+	mm := NewMsg(fmt.Sprintf(objMetaPreTmpl, info.Bucket, encodeName(info.ObjectMeta.Name)))
+	mm.Header.Set(MsgRollup, MsgRollupSubject)
+	mm.Data = data
+	if _, err := js.PublishMsg(mm); err != nil {
+		return err
+	}
+
+	// set the ModTime in case it's returned to the user, even though it's not the correct time.
+	info.ModTime = time.Now().UTC()
+	return nil
+}
+
+// AddLink will add a link to another object if it's not deleted and not another link
+// name is the name of this link object
+// obj is what is being linked too
 func (obs *obs) AddLink(name string, obj *ObjectInfo) (*ObjectInfo, error) {
-	if obj == nil {
-		return nil, errors.New("nats: object required")
+	if name == "" {
+		return nil, ErrNameRequired
+	}
+
+	// TODO Handle stale info
+
+	if obj == nil || obj.Name == "" {
+		return nil, ErrObjectRequired
 	}
 	if obj.Deleted {
-		return nil, errors.New("nats: object is deleted")
+		return nil, ErrNoLinkToDeleted
 	}
-	name = sanitizeName(name)
-	if !keyValid(name) {
-		return nil, ErrInvalidObjectName
+	if obj.isLink() {
+		return nil, ErrNoLinkToLink
 	}
 
-	// Same object store.
-	if obj.Bucket == obs.name {
-		info := *obj
-		info.Name = name
-		if err := obs.UpdateMeta(obj.Name, &info.ObjectMeta); err != nil {
-			return nil, err
+	// If object with link's name is found, error.
+	// If link with link's name is found, that's okay to overwrite.
+	// If there was an error that was not ErrObjectNotFound, error.
+	einfo, err := obs.GetInfo(name, GetObjectInfoShowDeleted())
+	if einfo != nil {
+		if !einfo.isLink() {
+			return nil, ErrObjectAlreadyExists
 		}
-		return obs.GetInfo(name)
+	} else if err != ErrObjectNotFound {
+		return nil, err
 	}
 
-	link := &ObjectLink{Bucket: obj.Bucket, Name: obj.Name}
+	// create the meta for the link
 	meta := &ObjectMeta{
 		Name: name,
-		Opts: &ObjectMetaOptions{Link: link},
+		Opts: &ObjectMetaOptions{Link: &ObjectLink{Bucket: obj.Bucket, Name: obj.Name}},
 	}
-	return obs.Put(meta, nil)
+	info := &ObjectInfo{Bucket: obs.name, NUID: nuid.Next(), ModTime: time.Now().UTC(), ObjectMeta: *meta}
+
+	// put the link object
+	if err = publishMeta(info, obs.js); err != nil {
+		return nil, err
+	}
+
+	return info, nil
 }
 
 // AddBucketLink will add a link to another object store.
 func (ob *obs) AddBucketLink(name string, bucket ObjectStore) (*ObjectInfo, error) {
+	if name == "" {
+		return nil, ErrNameRequired
+	}
 	if bucket == nil {
-		return nil, errors.New("nats: bucket required")
+		return nil, ErrBucketRequired
 	}
-	name = sanitizeName(name)
-	if !keyValid(name) {
-		return nil, ErrInvalidObjectName
-	}
-
 	bos, ok := bucket.(*obs)
 	if !ok {
-		return nil, errors.New("nats: bucket malformed")
+		return nil, ErrBucketMalformed
 	}
+
+	// If object with link's name is found, error.
+	// If link with link's name is found, that's okay to overwrite.
+	// If there was an error that was not ErrObjectNotFound, error.
+	einfo, err := ob.GetInfo(name, GetObjectInfoShowDeleted())
+	if einfo != nil {
+		if !einfo.isLink() {
+			return nil, ErrObjectAlreadyExists
+		}
+	} else if err != ErrObjectNotFound {
+		return nil, err
+	}
+
+	// create the meta for the link
 	meta := &ObjectMeta{
 		Name: name,
 		Opts: &ObjectMetaOptions{Link: &ObjectLink{Bucket: bos.name}},
 	}
-	return ob.Put(meta, nil)
+	info := &ObjectInfo{Bucket: ob.name, NUID: nuid.Next(), ObjectMeta: *meta}
+
+	// put the link object
+	err = publishMeta(info, ob.js)
+	if err != nil {
+		return nil, err
+	}
+
+	return info, nil
 }
 
 // PutBytes is convenience function to put a byte slice into this object store.
@@ -659,7 +801,7 @@ func (obs *obs) PutBytes(name string, data []byte, opts ...ObjectOpt) (*ObjectIn
 }
 
 // GetBytes is a convenience function to pull an object from this object store and return it as a byte slice.
-func (obs *obs) GetBytes(name string, opts ...ObjectOpt) ([]byte, error) {
+func (obs *obs) GetBytes(name string, opts ...GetObjectOpt) ([]byte, error) {
 	result, err := obs.Get(name, opts...)
 	if err != nil {
 		return nil, err
@@ -673,13 +815,13 @@ func (obs *obs) GetBytes(name string, opts ...ObjectOpt) ([]byte, error) {
 	return b.Bytes(), nil
 }
 
-// PutBytes is convenience function to put a string into this object store.
+// PutString is convenience function to put a string into this object store.
 func (obs *obs) PutString(name string, data string, opts ...ObjectOpt) (*ObjectInfo, error) {
 	return obs.Put(&ObjectMeta{Name: name}, strings.NewReader(data), opts...)
 }
 
 // GetString is a convenience function to pull an object from this object store and return it as a string.
-func (obs *obs) GetString(name string, opts ...ObjectOpt) (string, error) {
+func (obs *obs) GetString(name string, opts ...GetObjectOpt) (string, error) {
 	result, err := obs.Get(name, opts...)
 	if err != nil {
 		return _EMPTY_, err
@@ -704,7 +846,7 @@ func (obs *obs) PutFile(file string, opts ...ObjectOpt) (*ObjectInfo, error) {
 }
 
 // GetFile is a convenience function to pull and object and place in a file.
-func (obs *obs) GetFile(name, file string, opts ...ObjectOpt) error {
+func (obs *obs) GetFile(name, file string, opts ...GetObjectOpt) error {
 	// Expect file to be new.
 	f, err := os.OpenFile(file, os.O_WRONLY|os.O_CREATE, 0600)
 	if err != nil {
@@ -724,19 +866,54 @@ func (obs *obs) GetFile(name, file string, opts ...ObjectOpt) error {
 	return err
 }
 
+type GetObjectInfoOpt interface {
+	configureGetInfo(opts *getObjectInfoOpts) error
+}
+type getObjectInfoOpts struct {
+	ctx context.Context
+	// Include deleted object in the result.
+	showDeleted bool
+}
+
+type getObjectInfoFn func(opts *getObjectInfoOpts) error
+
+func (opt getObjectInfoFn) configureGetInfo(opts *getObjectInfoOpts) error {
+	return opt(opts)
+}
+
+// GetObjectInfoShowDeleted makes GetInfo() return object if it was marked as deleted.
+func GetObjectInfoShowDeleted() GetObjectInfoOpt {
+	return getObjectInfoFn(func(opts *getObjectInfoOpts) error {
+		opts.showDeleted = true
+		return nil
+	})
+}
+
+// For nats.Context() support.
+func (ctx ContextOpt) configureGetInfo(opts *getObjectInfoOpts) error {
+	opts.ctx = ctx
+	return nil
+}
+
 // GetInfo will retrieve the current information for the object.
-func (obs *obs) GetInfo(name string) (*ObjectInfo, error) {
-	// Lookup the stream to get the bound subject.
-	obj := sanitizeName(name)
-	if !keyValid(obj) {
-		return nil, ErrInvalidObjectName
+func (obs *obs) GetInfo(name string, opts ...GetObjectInfoOpt) (*ObjectInfo, error) {
+	// Grab last meta value we have.
+	if name == "" {
+		return nil, ErrNameRequired
+	}
+	var o getObjectInfoOpts
+	for _, opt := range opts {
+		if opt != nil {
+			if err := opt.configureGetInfo(&o); err != nil {
+				return nil, err
+			}
+		}
 	}
 
-	// Grab last meta value we have.
-	meta := fmt.Sprintf(objMetaPreTmpl, obs.name, obj)
+	metaSubj := fmt.Sprintf(objMetaPreTmpl, obs.name, encodeName(name)) // used as data in a JS API call
 	stream := fmt.Sprintf(objNameTmpl, obs.name)
 
-	m, err := obs.js.GetLastMsg(stream, meta)
+	m, err := obs.js.GetLastMsg(stream, metaSubj)
 	if err != nil {
 		if err == ErrMsgNotFound {
 			err = ErrObjectNotFound
@@ -747,29 +924,59 @@ func (obs *obs) GetInfo(name string) (*ObjectInfo, error) {
 	if err := json.Unmarshal(m.Data, &info); err != nil {
 		return nil, ErrBadObjectMeta
 	}
+	if !o.showDeleted && info.Deleted {
+		return nil, ErrObjectNotFound
+	}
 	info.ModTime = m.Time
 	return &info, nil
 }
 
-// UpdateMeta will update the meta data for the object.
+// UpdateMeta will update the meta for the object.
 func (obs *obs) UpdateMeta(name string, meta *ObjectMeta) error {
 	if meta == nil {
 		return ErrBadObjectMeta
 	}
-	// Grab meta info.
+
+	// Grab the current meta.
 	info, err := obs.GetInfo(name)
 	if err != nil {
+		if errors.Is(err, ErrObjectNotFound) {
+			return ErrUpdateMetaDeleted
+		}
 		return err
 	}
-	// Copy new meta
-	info.ObjectMeta = *meta
-	mm := NewMsg(fmt.Sprintf(objMetaPreTmpl, obs.name, sanitizeName(meta.Name)))
-	mm.Data, err = json.Marshal(info)
-	if err != nil {
+
+	// If the new name is different from the old, and it exists, error
+	// If there was an error that was not ErrObjectNotFound, error.
+	if name != meta.Name {
+		existingInfo, err := obs.GetInfo(meta.Name, GetObjectInfoShowDeleted())
+		if err != nil && !errors.Is(err, ErrObjectNotFound) {
+			return err
+		}
+		if err == nil && !existingInfo.Deleted {
+			return ErrObjectAlreadyExists
+		}
+	}
+
+	// Update Meta prevents update of ObjectMetaOptions (Link, ChunkSize)
+	// These should only be updated internally when appropriate.
+	info.Name = meta.Name
+	info.Description = meta.Description
+	info.Headers = meta.Headers
+
+	// Prepare the meta message
+	if err = publishMeta(info, obs.js); err != nil {
 		return err
 	}
-	_, err = obs.js.PublishMsg(mm)
-	return err
+
+	// did the name of this object change? We just stored the meta under the new name
+	// so delete the meta from the old name via purge stream for subject
+	if name != meta.Name {
+		metaSubj := fmt.Sprintf(objMetaPreTmpl, obs.name, encodeName(name))
+		return obs.js.purgeStream(obs.stream, &StreamPurgeRequest{Subject: metaSubj})
+	}
+
+	return nil
 }
 
 // Seal will seal the object store, no further modifications will be allowed.
@@ -864,21 +1071,71 @@ func (obs *obs) Watch(opts ...WatchOpt) (ObjectWatcher, error) {
 	return w, nil
 }
 
+type ListObjectsOpt interface {
+	configureListObjects(opts *listObjectOpts) error
+}
+type listObjectOpts struct {
+	ctx context.Context
+	// Include deleted objects in the result channel.
+	showDeleted bool
+}
+
+type listObjectsFn func(opts *listObjectOpts) error
+
+func (opt listObjectsFn) configureListObjects(opts *listObjectOpts) error {
+	return opt(opts)
+}
+
+// ListObjectsShowDeleted makes ListObjects() return deleted objects.
+func ListObjectsShowDeleted() ListObjectsOpt {
+	return listObjectsFn(func(opts *listObjectOpts) error {
+		opts.showDeleted = true
+		return nil
+	})
+}
+
+// For nats.Context() support.
+func (ctx ContextOpt) configureListObjects(opts *listObjectOpts) error {
+	opts.ctx = ctx
+	return nil
+}
+
 // List will list all the objects in this store.
-func (obs *obs) List(opts ...WatchOpt) ([]*ObjectInfo, error) {
-	opts = append(opts, IgnoreDeletes())
-	watcher, err := obs.Watch(opts...)
+func (obs *obs) List(opts ...ListObjectsOpt) ([]*ObjectInfo, error) {
+	var o listObjectOpts
+	for _, opt := range opts {
+		if opt != nil {
+			if err := opt.configureListObjects(&o); err != nil {
+				return nil, err
+			}
+		}
+	}
+	watchOpts := make([]WatchOpt, 0)
+	if !o.showDeleted {
+		watchOpts = append(watchOpts, IgnoreDeletes())
+	}
+	watcher, err := obs.Watch(watchOpts...)
 	if err != nil {
 		return nil, err
 	}
 	defer watcher.Stop()
+	if o.ctx == nil {
+		o.ctx = context.Background()
+	}
 
 	var objs []*ObjectInfo
-	for entry := range watcher.Updates() {
-		if entry == nil {
-			break
+	updates := watcher.Updates()
+Updates:
+	for {
+		select {
+		case entry := <-updates:
+			if entry == nil {
+				break Updates
+			}
+			objs = append(objs, entry)
+		case <-o.ctx.Done():
+			return nil, o.ctx.Err()
 		}
-		objs = append(objs, entry)
 	}
 	if len(objs) == 0 {
 		return nil, ErrNoObjectsFound
@@ -973,6 +1230,19 @@ func (o *objResult) Read(p []byte) (n int, err error) {
 			}
 		}
 	}
+	if err == io.EOF {
+		// Make sure the digest matches.
+		sha := o.digest.Sum(nil)
+		rsha, decodeErr := DecodeObjectDigest(o.info.Digest)
+		if decodeErr != nil {
+			o.err = decodeErr
+			return 0, o.err
+		}
+		if !bytes.Equal(sha[:], rsha) {
+			o.err = ErrDigestMismatch
+			return 0, o.err
+		}
+	}
 	return n, err
 }
 
@@ -1003,3 +1273,86 @@ func (o *objResult) Error() error {
 	defer o.Unlock()
 	return o.err
 }
+
+// ObjectStoreNames is used to retrieve a list of bucket names
+func (js *js) ObjectStoreNames(opts ...ObjectOpt) <-chan string {
+	var o objOpts
+	for _, opt := range opts {
+		if opt != nil {
+			if err := opt.configureObject(&o); err != nil {
+				return nil
+			}
+		}
+	}
+	ch := make(chan string)
+	var cancel context.CancelFunc
+	if o.ctx == nil {
+		o.ctx, cancel = context.WithTimeout(context.Background(), defaultRequestWait)
+	}
+	l := &streamLister{js: js}
+	l.js.opts.streamListSubject = fmt.Sprintf(objAllChunksPreTmpl, "*")
+	l.js.opts.ctx = o.ctx
+	go func() {
+		if cancel != nil {
+			defer cancel()
+		}
+		defer close(ch)
+		for l.Next() {
+			for _, info := range l.Page() {
+				if !strings.HasPrefix(info.Config.Name, "OBJ_") {
+					continue
+				}
+				select {
+				case ch <- info.Config.Name:
+				case <-o.ctx.Done():
+					return
+				}
+			}
+		}
+	}()
+
+	return ch
+}
+
+// ObjectStores is used to retrieve a list of bucket statuses
+func (js *js) ObjectStores(opts ...ObjectOpt) <-chan ObjectStoreStatus {
+	var o objOpts
+	for _, opt := range opts {
+		if opt != nil {
+			if err := opt.configureObject(&o); err != nil {
+				return nil
+			}
+		}
+	}
+	ch := make(chan ObjectStoreStatus)
+	var cancel context.CancelFunc
+	if o.ctx == nil {
+		o.ctx, cancel = context.WithTimeout(context.Background(), defaultRequestWait)
+	}
+	l := &streamLister{js: js}
+	l.js.opts.streamListSubject = fmt.Sprintf(objAllChunksPreTmpl, "*")
+	l.js.opts.ctx = o.ctx
+	go func() {
+		if cancel != nil {
+			defer cancel()
+		}
+		defer close(ch)
+		for l.Next() {
+			for _, info := range l.Page() {
+				if !strings.HasPrefix(info.Config.Name, "OBJ_") {
+					continue
+				}
+				select {
+				case ch <- &ObjectBucketStatus{
+					nfo:    info,
+					bucket: strings.TrimPrefix(info.Config.Name, "OBJ_"),
+				}:
+				case <-o.ctx.Done():
+					return
+				}
+			}
+		}
+	}()
+
+	return ch
+}
